package andante;

import com.ibm.aglet.*;
import com.ibm.aglet.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;

import org.omg.PortableServer.*;
import org.omg.CORBA.ORB;
import org.omg.CosNaming.*;
import org.omg.CosNaming.NamingContextPackage.*;

import andante.orb.*;
import andante.orb.stage.*;
import andante.orb.agent.*;
import andante.orb.devices.*;


public class Maestro extends MusicalAglet {

  static POA Poa;
  static ORB Orb;

  private MaestroAgentManager Agents;

  private SortedSet Score;
  private Metronome M;
  private int PulseCount = 0;

  private boolean Start = false;

  public void init(Stage s) {
  }

  public void run(BufferedReader br, Metronome m) {
    Agents = new MaestroAgentManager();
    M = m;
    parseScript(br);

    synchronized (this) {
      while (!Start) {
        try {
          wait();
        }
        catch (InterruptedException ie) {}
      }
    }

    try {
      MusicalAgentImpl ma = new MusicalAgentImpl(this);
      org.omg.CORBA.Object obj = Poa.servant_to_reference(ma);
      M.register(MusicalAgentHelper.narrow(obj));
    }
    catch (Exception e) {}
  }

  public void pulse() {
    System.out.println(PulseCount++);
  }

  private void parseScript(BufferedReader br) {
    MusicalAglet mAglet = null;

    int lineNum = 0;
    try {
      String line;
      StringTokenizer st;
      String token;
      String ident;

      line = br.readLine().trim(); lineNum++;
      while (line.equals("")) {
        line = br.readLine().trim(); lineNum++;
      }

      Hashtable args = new Hashtable();
      org.omg.CORBA.Object obj = Poa.servant_to_reference(Agents);
      MusicalAgentProxy map = MusicalAgentProxyHelper.narrow(obj);
      args.put("Proxy", map);

      List maglets = new ArrayList();

      // Declarations
      System.out.println("*parser: DEC");

      while (!line.equals("--")) {
        System.out.println("*parser: " +lineNum + ". dec: " + line);
        st = new StringTokenizer(line);
        String agentClass = st.nextToken();
        ident = st.nextToken();
        AgletProxy maProxy = AgletsServer.getContext().createAglet
          (null, agentClass, args);
        mAglet = (MusicalAglet)maProxy.getAglet();
        System.out.println("*parser: " +lineNum + ". c: " + ident + " " + mAglet.id());
        maglets.add(mAglet);
        MusicalAgentImpl mai = new MusicalAgentImpl(mAglet);
        obj = Poa.servant_to_reference(mai);
        MusicalAgent ma = MusicalAgentHelper.narrow(obj);
        Agents.put(ident, mAglet.getAgletID().toString(), ma);
        line = br.readLine().trim(); lineNum++;
        while (line.equals("")) {
          line = br.readLine().trim(); lineNum++;
          System.out.println("*parser: " +lineNum + "a. " + line);
        }
      }

      Iterator it = maglets.iterator();
      while (it.hasNext()) {
        MusicalAglet ma = (MusicalAglet)it.next();
        synchronized (ma) {
          while (!ma.ready()) {
            try {
              ma.wait();
            }
            catch (InterruptedException ie) {}
          }
        }
      }
      System.out.println("*parser: " +"Maestro: Agents ready");

      line = br.readLine().trim(); lineNum++;
      while (line.equals("")) {
        line = br.readLine().trim(); lineNum++;
      }

      // Initializations
      System.out.println("*parser: INIT");

      MusicalAgent ma;
      List params = new ArrayList();
      while (!line.equals("--")) {
        params.clear();
        st = new StringTokenizer(line);
        ident = st.nextToken();
        String message = st.nextToken().trim();
        ma = Agents.get(ident);
        while (st.hasMoreElements()) {
          params.add(st.nextToken());
        }
        Class[] paramTypes = new Class[params.size()];
        for (int i = 0; i < paramTypes.length; i++)
          paramTypes[i] = Class.forName("java.lang.String");
        Object[] paramsArray = params.toArray();
        System.out.print(ident + " " + message);
        for (int i = 0; i < paramTypes.length; i++)
          System.out.print(" " + paramsArray[i]);
        System.out.println();
        Method method = mAglet.getClass().getMethod(message, paramTypes);
        method.invoke(mAglet,paramsArray);
        line = br.readLine().trim(); lineNum++;
        while (line.equals("")) {
          line = br.readLine().trim(); lineNum++;
        }
      }

      line = br.readLine().trim(); lineNum++;
      while (line.equals("")) {
        line = br.readLine().trim(); lineNum++;
      }

      // Score
      System.out.println("*parser: SCORE");

      Score = new TreeSet();
      while (!line.equals("--")) {
        params.clear();
        st = new StringTokenizer(line);
        int time = Integer.parseInt(st.nextToken());
        ident = st.nextToken();
        String message = st.nextToken().trim();
        while (st.hasMoreElements()) {
          params.add(st.nextToken());
        }
        Class[] paramTypes = new Class[params.size()];
        for (int i = 0; i < paramTypes.length; i++)
          paramTypes[i] = "foo".getClass();
        Object[] paramsArray = params.toArray();

        Score.add(new EventHolder(time, ident, message, paramTypes, paramsArray));

        line = br.readLine().trim(); lineNum++;
        while (line.equals("")) {
          line = br.readLine().trim(); lineNum++;
        }
      }

      it = Score.iterator();
      while (it.hasNext()) {
        EventHolder eh = (EventHolder)it.next();
        System.out.println("*parser: " +eh);
      }

      synchronized (this) {
        Start = true;
        notifyAll();
      }

//      Agents.print(System.out);
      
    }
    catch (NoSuchElementException nsee) { 
      nsee.printStackTrace();
      throw new Error("Error while parsing the script (l." + lineNum + ")");
    }
    catch (Exception e) { 
      System.out.println("*parser: " +"Unhandled exception! (l." + lineNum + ")");
      e.printStackTrace();
    }
  }

  class EventHolder implements Comparable {
    private int TimeStamp;
    private String Ident;
    private String Message;
    private Class [] ParamTypes;
    private Object [] Params;

    public EventHolder(int t, String id, String m, Class [] pt, Object [] p) {
      TimeStamp = t;
      Ident = id;
      Message = m;
      ParamTypes = pt;
      Params = p;
    }
    
    public final int timeStamp() { return TimeStamp; }
    public final String ident() { return Ident; }
    public final String message() { return Message; }
    public final Class [] paramTypes() { return ParamTypes; }
    public final Object [] params() { return Params; }

    public int compareTo(Object other) {
      EventHolder eh = (EventHolder)other;
      return timeStamp() - eh.timeStamp();
    }
    public String toString() {
      return new String(TimeStamp + " " + Ident + " " + Message);
    }
  }

  public static void main(String []args) {
    try {
      Properties props = new Properties();
      props.put("org.omg.CORBA.ORBInitialHost", "methos");
      props.put("org.omg.CORBA.ORBInitialPort", "1050");
      Orb = org.omg.CORBA.ORB.init(args, props);
      
      AudioDeviceImpl daI = new AudioDeviceImpl();
      AudioDevice da = null;
      StageImpl stageI = null;
      Stage stage = null;

      AgletsServer.main(args);

      NamingContextExt nc = NamingContextExtHelper.narrow
	(Orb.resolve_initial_references("NameService"));

      NamingContext nc2 = null;
      try {
	nc2 = NamingContextHelper.narrow(nc.resolve(nc.to_name("andante")));
      } catch (NotFound nf) {
	System.out.println(nf);
	nc2 = nc.bind_new_context(nc.to_name("andante"));
      }

      String StageName = InetAddress.getLocalHost().getHostName() + ":" +
	AgletsServer.getPortNum();

      StageName = StageName.replace('.', '-');
      System.out.println("Stage name: " + StageName);

      NamingContext LocalNC = null;
      try {
	LocalNC = NamingContextHelper.narrow(nc2.resolve(nc.to_name(StageName)));
      } catch (NotFound nf) {
	System.out.println(nf);
	LocalNC = nc2.bind_new_context(nc.to_name(StageName));
      }


      Poa = org.omg.PortableServer.POAHelper.narrow
	( Orb.resolve_initial_references("RootPOA"));
      Poa.the_POAManager().activate();
      org.omg.CORBA.Object obj = Poa.servant_to_reference(daI);
      da = AudioDeviceHelper.narrow(obj);

      stageI = new StageImpl(StageName, da, Poa);

      obj = Poa.servant_to_reference(stageI);
      stage = StageHelper.narrow(obj);

      NameComponent [] NCName = nc.to_name("Metronome");
      LocalNC.rebind(NCName, stage.getMetronome());

      AgletsServer.getContext().setProperty("Stage", stage);
      AgletsServer.getContext().setProperty("Orb", Orb);
      AgletsServer.getContext().setProperty("NamingContext", LocalNC);

      AgletProxy proxy = AgletsServer.getContext().createAglet
	(null, "andante.Maestro", null);
      Maestro m = (Maestro)proxy.getAglet();
//      Maestro m = new Maestro();
      proxy.dispatch(new URL("atp://localhost:4445"));

      BufferedReader br = new BufferedReader(new FileReader(args[args.length-1]));

      (new Thread() {
        public void run() {
          Orb.run();
        }
        }).start();

      m.run(br, stageI.getMetronome());
    }
    catch (Exception e) { e.printStackTrace(); }
  }
}

  class MaestroAgentManager extends MusicalAgentProxyPOA implements Serializable {
    private Map GIds; // Id -> AgletId
    private Map Servants; // AgletId -> MusicalAgent

    public MaestroAgentManager() {
      Servants = new Hashtable();
      GIds = new Hashtable();
    }

    public synchronized void put(String id, String gid, MusicalAgent ma) {
      System.out.println("Manager: put: " + id + " " + gid);
      GIds.put(id, gid);
      Servants.put(gid, ma);
    }

    public synchronized MusicalAgent get(String id) {
      return (MusicalAgent)Servants.get(GIds.get(id));
    }

    public synchronized void setMusicalAgent(String gid, MusicalAgent ma) {
//      System.out.println("     Maestro: setMusicalAgent 1 " + gid);
//      System.out.println("     Maestro: setMusicalAgent 2 " + Servants);
//      System.out.println("     Maestro: setMusicalAgent 3 " + ma);
      System.out.println("     Maestro: setMusicalAgent 4 " + ma.name());

      System.out.println("Manager: set: " + gid);
      Servants.put(gid, ma);
    }

    public synchronized void foo(String gid) {
      System.out.println("     Maestro: foo 1 " + gid);
    }

    public synchronized void print(PrintStream ps) {
      Iterator it = GIds.keySet().iterator();

      while (it.hasNext()) {
        String id = (String)it.next();
        ps.println(id + " -> " + GIds.get(id) +
                   " -> " + ((MusicalAgent)Servants.get(GIds.get(id))).name());
      }
    }

  }

