package andante.apps.noiseweaver;

import com.ibm.aglet.*;
import com.ibm.aglet.event.*;
import java.io.*;
import java.net.*;

import andante.agent.*;

import andante.orb.*;
import andante.orb.stage.*;
import andante.orb.agent.*;
import java.util.*;

public class NoiseAgent extends MusicalAglet {

  private NumbersGenerator NGPitch;
  private NumbersGenerator NGIntensity;
  private NumbersGenerator NGDuration;
  private Channel Channel;

  private int[] ScaleMapping;

  private Stage St;

  public void onCreation (Object o) {
    super.onCreation(o);

    Name = "NoiseAgent " + ++Cont;

    Props.setProperty("PitchGenerator", "Pink");
    Props.setProperty("IntensityGenerator", "Pink");
    Props.setProperty("IntensityMax", "110");
    Props.setProperty("IntensityMin", "80");
    Props.setProperty("DurationGenerator", "Pink");
    Props.setProperty("DurationMax", "5");
    Props.setProperty("DurationMin", "1");
    Props.setProperty("Channel", "1");
    Props.setProperty("Instrument", "1");
    Props.setProperty("Scale", "Diatonic");
    Props.setProperty("ScaleStart", "60");
    Props.setProperty("ScaleLength", "14");
    
    NGPitch = newNumbersGenerator(Props.getProperty("PitchGenerator"),
				  0,
				  Integer.parseInt(Props.getProperty("ScaleLength"))
				  
      );
    NGIntensity = newNumbersGenerator(Props.getProperty("IntensityGenerator"),
				     Integer.parseInt(Props.getProperty("IntensityMin")),
				     Integer.parseInt(Props.getProperty("IntensityMax"))
      );
    NGDuration = newNumbersGenerator(Props.getProperty("DurationGenerator"),
				     Integer.parseInt(Props.getProperty("DurationMin")),
				     Integer.parseInt(Props.getProperty("DurationMax"))
      );

    ScaleMapping = Scales.getScale(Props.getProperty("Scale"),
				   Integer.parseInt(Props.getProperty("ScaleStart")),
				     Integer.parseInt(Props.getProperty("ScaleLength"))
      );

  }

  public void setProperty(String p, String v) {
    if (!Props.keySet().contains(p))
      return;

    System.out.println("Set property " + p + " to " + v);

    Props.setProperty(p, v);

    if (p.startsWith("Pitch")) {
      NGPitch = newNumbersGenerator(Props.getProperty("PitchGenerator"),
				    0,
				    Integer.parseInt(Props.getProperty("ScaleLength"))
				    
	);
    } else if (p.startsWith("Intensity")) {
      NGIntensity = newNumbersGenerator(Props.getProperty("IntensityGenerator"),
				       Integer.parseInt(Props.getProperty("IntensityMin")),
				       Integer.parseInt(Props.getProperty("IntensityMax"))
	);
    } else if (p.startsWith("Duration")) {
      NGDuration = newNumbersGenerator(Props.getProperty("DurationGenerator"),
				       Integer.parseInt(Props.getProperty("DurationMin")),
				    Integer.parseInt(Props.getProperty("DurationMax"))
	);
    } else if (p.equals("Channel")) {
      Channel = St.channel(Integer.parseInt(Props.getProperty("Channel")));
    } else if (p.equals("Instrument")) {
      Channel.changeInstrument(Short.parseShort(Props.getProperty("Instrument")));
    } else if (p.startsWith("Scale")) {
      ScaleMapping = Scales.getScale(Props.getProperty("Scale"),
				     Integer.parseInt(Props.getProperty("ScaleStart")),
				     Integer.parseInt(Props.getProperty("ScaleLength"))
	);
      if (p.equals("ScaleLength")) {
	NGPitch = newNumbersGenerator(Props.getProperty("PitchGenerator"),
				      0,
				      Integer.parseInt(Props.getProperty("ScaleLength"))
				      
	  );
      }
    }
  }

  private NumbersGenerator newNumbersGenerator(String name, int min, int max) {
    NumbersGenerator res = null;
    if (name.equals("White")) {
      res = new WhiteNoise(min, max);
    } else if (name.equals("Pink")) {
      res = new PinkNoise(min, max);
    } else if (name.equals("Brownian")) {
      res = new BrownianNoise(min, max);
    } else if (name.equals("Constant")) {
      res = new ConstantNG(min, max);
    } else {
      res = new WhiteNoise(min, max);      
    }
    return res;
  }

  public void init(Stage stage) {
    
    St = stage;

    Channel = St.channel(
      Short.parseShort(Props.getProperty("Channel")));

    Channel.changeInstrument(Short.parseShort(Props.getProperty("Instrument")));
    
    M = St.getMetronome();

    System.out.println("init: " + this);
    System.out.println("init: " + M);
/*    while (true) {
      short a = (short)ScaleMapping[(NGPitch.next()) % ScaleMapping.length]; // FIXME!!!!
      short i = (short)(NGIntensity.next());
      Channel.noteOn(a, i);
      try { Thread.sleep(NGDuration.next() * 10); }
      catch (InterruptedException e) {}
      Channel.noteOff(a, i);
    }
*/

  }


  public void play() {
    Playing = true;
    System.out.println("play: " + this);
    System.out.println("play: " + M);
    M.register(ThisServant);
  }
  public void stop() {
    M.unregister(ThisServant);
    Playing = false;
  }

  short _dur = 0;
  public void pulse() {
    if (_dur == 0) {
      andante.orb.Note n = new andante.orb.Note();
      n.pitch = (short)ScaleMapping[(NGPitch.next()) % ScaleMapping.length];
      n.intensity = (short)(NGIntensity.next());
      _dur = (short)NGDuration.next();
      n.duration = (short)(_dur * M.getPulseMillis());
      Channel.playNote(n);
    }
    _dur--;
  }

}
